import 'dart:ffi';
import 'package:ffi_helper/src/array_utils.dart';
import 'package:meta/meta.dart';
import 'package:ffi/ffi.dart' as ffi;

import 'array.dart';

/// Instead of doing many little allocations, [Arena]
/// allocates a large chunk of memory which can be sliced into multiple [ArrayView]s.
/// This can significantly improve performance and it makes
/// memory management simpler.
/// ````
/// final arena = Arena<Uint8>.allocate(10);
/// final array = arena.newArray(5);
/// final array2 = arena.newArray(5);
/// void some_native_c_api(array.ptr, array.length, array2.ptr, array2.length);
/// arena.delete();
/// ````
class Arena<T extends NativeType> {
  int _index = 0;

  int get length => buffer.length;

  @protected
  final Array<T> buffer;

  /// Left free space of Arena.
  int get freeSpaceLeft => length - _index;

  Arena.fromBuffer(this.buffer);
  Arena(int length, [Allocator allocate = ffi.allocate])
      : buffer = Array<T>(length, allocate);

  /// Returns a view into [buffer].
  /// When [length] exceeds [freeSpaceLeft] it throws a [RangeError].
  @mustCallSuper
  ArrayView<T> newArray(int length) {
    assert(length > 0);
    if (freeSpaceLeft < length) {
      throw RangeError('Not enough free space');
    }
    final ptrAtIndex = buffer.ptr.elementAt(_index);
    final arr = ArrayView<T>.fromPtr(ptrAtIndex, length);
    _index += length;
    return arr;
  }

  void delete() {
    buffer.delete();
  }
}
