import 'array_utils.dart';
import 'package:meta/meta.dart';
import 'array.dart';
import 'dart:ffi';
import 'package:ffi/ffi.dart' as ffi;

/// Reference counted Array. All references share the same buffer,
/// which gets deleted when all references are deleted.
class SharedArray<T extends NativeType> {
  int _refCount = 0;
  int get refCount => _refCount;
  @doNotStore
  Array<T> sharedBuffer;
  bool _calledDelete = false;

  SharedArray.fromBuffer(this.sharedBuffer);
  SharedArray(int length, [Allocator allocate = ffi.allocate])
      : sharedBuffer = Array<T>(length, allocate);

  /// Create new reference to this SharedArray.
  @mustCallSuper
  ArrayRef<T> newRef() {
    if (_calledDelete) {
      throw StateError('Cannot create new ArrayRef to deleted SharedArray');
    }
    _refCount++;
    return ArrayRef._(this);
  }

  /// After calling [delete], the [sharedBuffer] will get deleted when there are no more
  /// dangling [ArrayRef]s.
  void delete() {
    if (_calledDelete) return;
    _calledDelete = true;
    if (refCount == 0) {
      sharedBuffer.delete();
    }
  }
}

/// Reference to a [SharedArray].
/// Not calling [delete] leads to memory leak.
class ArrayRef<T extends NativeType> implements Array<T> {
  final SharedArray<T> _shared;
  @override
  @doNotStore
  Pointer<T> get ptr => _shared.sharedBuffer.ptr;
  @override
  int get length => _shared.sharedBuffer.length;
  bool _freed = false;

  ArrayRef._(this._shared);

  /// Reduces reference count.
  @override
  @mustCallSuper
  void delete() {
    if (_freed) {
      return;
    }
    assert(_shared._refCount > 0);
    _shared._refCount--;
    if (_shared._refCount == 0 && _shared._calledDelete) {
      _shared.sharedBuffer.delete();
    }
    _freed = true;
  }

  @override
  bool operator ==(dynamic arr) {
    return _shared.sharedBuffer == arr;
  }

  @override
  int get hashCode => _shared.sharedBuffer.hashCode;

  @override
  String toString() {
    return _shared.sharedBuffer.toString();
  }
}
